/*
 *  HSMultiResol.cpp
 *
 *
 *  Created by Patrick Heas on 02/12/12.
 *  Copyright 2012 INRIA. All rights reserved.
 *
 */
#include <iostream>
#include <string> 
#include <fstream>
#include <cstdio> 
#include <cstdlib> 
#include <cmath>
#include <cassert>

#include "routinesHSMultiResol.h" 
#include "CImg.h"
#include "../Lib/Corpetti/io_field.h"

#include      "../Lib/Spline/coeff.h"
#include      "../Lib/Spline/interpol.h"
#include      "../Lib/Spline/derivee_spline.h"
#include      "../Lib/Spline/interpol.h"
#include      "../Lib/Spline/io.h"

#define ABS(n)   ((n > 0) ? (n) : -(n))
#define Max(n,m)   ((n > m) ? (n) : m)

using namespace cimg_library;
using namespace std;

/*************************************************************************************************************/

/*************************************************************************************************************/
/**********************************Spline Interpolation  and Derivative Filters ******************************/
/*************************************************************************************************************/

/*************************************************************************************************************/
void Recal_spline(CImg<double> &out,const CImg<double> &in,const CImg<double> &UV, int SplineDegree){

  cout<<"\n Recalage"<<endl;
  short int nb_col=in.width();
  short int nb_lig=in.height();
  float *Im_in;
  float *Im_out=new float[nb_col*nb_lig];
  float *u;
  float *v;
	
  Im_in=CImg2float(in);
  u=CImg2float(UV.get_slice(0));
  v=CImg2float(UV.get_slice(1));

  float	*ImageRasterArray=(float*)calloc(nb_col*nb_lig,sizeof(float));
  Copy_f(ImageRasterArray,Im_in,nb_col,nb_lig);
  if (SplineDegree>2){bool Error = SamplesToCoefficients(ImageRasterArray, nb_col, nb_lig, SplineDegree); // on recupere le coefficient
    if(Error){printf("probleme\n");}
  }

  short i,j;
  int I_u,I_v;
  float Du,Dv;
  float U,V;

  float *ad;
  float p00,p01,p10,p11;
  int D;
  short inc_i,inc_j;
  float pix;float * ptImout;float * ptImin;float * ptU;float * ptV;
  ptImout=Im_out;ptImin=Im_in;ptU=u;ptV=v;

  for(i=0;i<nb_lig;i++)
    for(j=0;j<nb_col;j++,ptImin++,ptImout++,ptU++,ptV++){
      U=(*ptU);
      V=(*ptV);
      if(U>0)
	inc_j=1;
      else
	inc_j=-1;

      if(V>0)
	inc_i=1;
      else
	inc_i=-1;

      I_u=(int)(U);
      I_v=(int)(V);

      Du=ABS(U)-ABS(I_u);
      Dv=ABS(V)-ABS(I_v);

      if((float)j+U<0.0)
	{ I_u=-j;
	  Du=0.0;
	}

      if((float)j+U >(float)(nb_col-1))
	{ I_u=nb_col-1-j;
	  Du=0.0;
	}

      if( (float)i+V<0.0)
	{	I_v=-i;
	  Dv=0.0;
	}

      if((float)i+V>(float)(nb_lig-1))
	{	I_v=nb_lig-1-i;
	  Dv=0.0;
	}

      D=I_u + (I_v * nb_col);
      ad=ptImin + D;

      p11=Du*Dv;
      p01=Du*(1-Dv);
      p10=(1-Du)*Dv;
      p00=(1-Du)*(1-Dv);

      /* -- bords images --*/
      if((I_u+j+inc_j)>=nb_col || (I_u+j+inc_j)<0 || (I_v+i+inc_i)>=nb_lig || (I_v+i+inc_i)<0){
	pix = (*ad);
      }
      /* -- interieur images --*/
      else{
	if(abs(in(I_u+j+inc_j,I_v+i+inc_i))>=1022||abs(in(I_u+j,I_v+i))>=1022||abs(in(I_u+j+inc_j,I_v+i))>=1022||abs(in(I_u+j,I_v+i+inc_i))>=1022
	   /*||(in(I_u+j+inc_j,I_v+i+inc_i))==0||(in(I_u+j,I_v+i))==0||(in(I_u+j+inc_j,I_v+i))==0||(in(I_u+j,I_v+i+inc_i))==0*/){
	  pix = (*ad);
	}
	else{
	  if (SplineDegree>2){
	    pix = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+U, i+V, SplineDegree);
	  }
	  else{   
	    pix = (*ad)*p00 +
	      (*(ad+inc_j))*p01 +
	      (*(ad+inc_i*nb_col))*p10 +
	      (*(ad+inc_j+inc_i*nb_col))*p11;
	  }
	}
      }
      *ptImout= pix;
    }
  out=float2CImg(out.width(),out.height(), Im_out);
  free(ImageRasterArray);
  delete [] Im_in;delete [] Im_out;delete [] u;delete [] v;
}
void Proj_Dup_Spline(CImg<double> &out, const CImg<double> &in, int SplineDegree){

	short int nb_col=in.width();
	short int nb_lig=in.height();
	float *I_low;
	float *I_high=new float[nb_col*nb_lig*4];
	I_low=CImg2float(in); 
	float	*ImageRasterArray=(float*)calloc(nb_col*nb_lig,sizeof(float));
	Copy_f(ImageRasterArray,I_low,nb_col,nb_lig);
	bool Error = SamplesToCoefficients(ImageRasterArray, nb_col, nb_lig, SplineDegree); // on recupere le coefficient
	if(Error){ printf("probleme\n"); }
	
    	float *pt;
   	short L_s;
   	short i,j;
	int count=0;

   	 L_s=nb_col<<1;
   	 pt=I_high;    
    
  	  for(i=0;i<(nb_lig-1);i++,pt+=L_s,count+=L_s)
	  {
 	   	for(j=0;j<(nb_col-1);j++,pt+=2,count+=2)
 	   	{
	   	*(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree);
	   	*(pt+L_s) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i+0.5, SplineDegree);	
	   	*(pt+1) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+0.5, i, SplineDegree);
	   	*(pt+1+L_s) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+0.5, i+0.5, SplineDegree);
 	   	}
		//bord droit
	   	*(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree); 
		*(pt+L_s) =(float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i+0.5 , SplineDegree);
		*(pt+1) = *(pt);
	   	*(pt+1+L_s) =  *(pt+L_s);
		pt+=2;count+=2;
	   }
	   //derniere ligne
	  for(j=0;j<(nb_col-1);j++,pt+=2,count+=2)
 	   {		
	   *(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree);
	   *(pt+L_s) = *(pt);	
	   *(pt+1) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+0.5, i, SplineDegree);
	   *(pt+1+L_s) = *(pt+1);
 	   }
	   //coin droite-bas
	   *(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree);
	   *(pt+L_s) = *(pt);	
	   *(pt+1) = *(pt);
	   *(pt+1+L_s) = *(pt);
	   out=float2CImg(out.width(),out.height(), I_high);
	   free(ImageRasterArray);
	   delete [] I_high;delete [] I_low;
	   
}
CImg<double> gradSpline_X(const CImg<double> &in, float sigma){

  int SplineDegree=3;
  short int nb_col=in.width();
  short int nb_lig=in.height();
  CImg<double> out(nb_col,nb_lig);
  CImg<double> tmp_1(nb_col,nb_lig);
  CImg<double> tmp0(nb_col,nb_lig);
  CImg<double> tmp1(nb_col,nb_lig);
  float *u;
  float *coef=(float*)calloc(nb_col*nb_lig, sizeof(float));
  if (sigma!=0){out=in.get_blur(sigma,true);}else{out=in;}
  u=CImg2float(out);
  if (directBsplineMirror(u,coef, nb_col, nb_lig, 1,SplineDegree ) != 0){exit(1);}
  gradX(coef,u,nb_col,nb_lig, SplineDegree,0); 
  out=float2CImg(out.width(),out.height(), u);
  free(coef);
  delete [] u;
  return out;

}
CImg<double> gradSpline_Y(const CImg<double> &in, float sigma){

  int SplineDegree=3;
  short int nb_col=in.width();
  short int nb_lig=in.height();
  CImg<double> out(nb_col,nb_lig);
  CImg<double> tmp_1(nb_col,nb_lig);
  CImg<double> tmp0(nb_col,nb_lig);
  CImg<double> tmp1(nb_col,nb_lig);
  float *u;
  float *coef=(float*)calloc(nb_col*nb_lig, sizeof(float));
  if (sigma!=0){out=in.get_blur(sigma,true);}else{out=in;}
  u=CImg2float(out);
  if (directBsplineMirror(u,coef, nb_col, nb_lig, 1,SplineDegree ) != 0){exit(1);}
  gradY(coef,u,nb_col,nb_lig, SplineDegree,0); 
  out=float2CImg(out.width(),out.height(), u);
  free(coef);
  delete [] u;
  return out;

}


CImg<double> grad_t(const CImg<double> &in,const CImg<double> &indt, float sigma){
	
  short int nb_col=in.width();
  short int nb_lig=in.height();
  CImg<double> out(nb_col,nb_lig);

  if (sigma!=0){
    out=(indt-in).get_blur(sigma,true);
  }
  else{out=indt-in;}
  return out;	
}

/*************************************************************************************************************/

/*************************************************************************************************************/
/********************************** Ecriture champs    *************************************/
/*************************************************************************************************************/

/*************************************************************************************************************/

void write_aMotionFile(char *fileMotionOut, const CImg<double>& im){
	
	int dx=im.width();
   	int dy=im.height();
  	float *ima_u=new float[dx*dy];
	float *ima_v=new float[dx*dy];
	for (int j=0;j<dy;j++){
		 for (int i=0;i<dx;i++){
			ima_u[j*dx+i]=im(i,j,0);
			ima_v[j*dx+i]=im(i,j,1);
	 	 }
	}
	cout<<fileMotionOut<<endl;
	write_champ(ima_u, ima_v, dx, dy,fileMotionOut);
	delete [] ima_u;
	delete [] ima_v;

}
CImg<float> read_aMotionFile(char * filename,int dx,int dy,int Dx, int ofx, int ofy){

	int ii,jj;
	float u,v;
	FILE * file=NULL;
	int readInt;
	char * readLine =new char[200];
	 
	CImg<float> im(dx,dy,2);
	CImg<float> aux(Dx,dy+ofy, 2);

	
	assert (file=fopen(filename,"r"));
	cout<<filename<<endl;
	fscanf (file,"%d %d\n",&readInt,&readInt);		
	fgets(readLine,100,file);
	for (int j=0;j<dy+ofy;j++){
		for (int i=0;i<Dx;i++){
			fscanf(file,"%d %d %f %f\n",&ii,&jj,&u,&v);
			aux(i,j,0)=u;
			aux(i,j,1)=v;
		}
	}
  	fclose(file);
	for (int j=0;j<dy;j++){
		for (int i=0;i<dx;i++){
			im(i,j,0)=aux(i+ofx,j+ofy,0);
			im(i,j,1)=aux(i+ofx,j+ofy,1);
		}
	}

	delete [] readLine;

	return 	im;
}

/*************************************************************************************************************/

/*************************************************************************************************************/
/**************************************************** Main   *************************************************/
/*************************************************************************************************************/

/*************************************************************************************************************/

int main(int argc, char *argv [])
{  
  cimg_usage("Optic flow estimation with Horn Schunck and optimal regularization coef estimation\n (Patrick Heas, dec. 2012)\n All right reserved");
	
  /*-- Parameters --*/ 
  string ifn = cimg_option("-inPath","inputHS","file of 2 input files in pgm format");	
  int dy = cimg_option("-dy",256,"tile line number");
  int dx  = cimg_option("-dx",256,"tile column number");
  int ofy = cimg_option("-ofy",0,"offset line number ");
  int ofx  = cimg_option("-ofx",0,"offset column number");
  int Dx  = cimg_option("-Dx",256,"whole Image column number");
  int Dy  = cimg_option("-Dy",256,"whole Image line number");
 	
  double norm= cimg_option("-norm",2.,"Norm data term : 2=L2,1=L1,0=Leclerc (default=L2)");
  double normRegRob= cimg_option("-normRegRob",2,"Norm reg term : 2=L2,1=L1,0=Leclerc (default=L2)");
  double sigma_obs= cimg_option("-sigma_obs",50.,"Initial robust parameter control on data term (norm leclerc)");
  double sigma_reg= cimg_option("-sigma_reg",50.,"Initial robust parameter control on reg term (norm leclerc)");
  bool sigmaRobEstim  = cimg_option("-sigmaRobEstim",false,"bayesian estimation of robust parameter observation (default=false) ");
  bool sigmaRegRobEstim   = cimg_option("-sigmaRegRobEstim",false,"bayesian estimation of robust parameter regularisation (default=false) ");
  if (norm==2&&normRegRob==2){sigmaRobEstim=false;sigmaRegRobEstim=false;}
  bool sigmaAlphaEstim  = cimg_option("-sigmaAlphaEstim",false,"bayesian estimation of alpha reg parameter (default=true) ");
  bool sigmaBetaEstim  = cimg_option("-sigmaBetaEstim",false,"bayesian estimation of beta reg parameter (default=true) ");
  
  double sigBlur = cimg_option("-sigBlur",0.,"Amplitude Tchumperlé's isotropic blurring");
  int nb_resol  = cimg_option("-multiresol",3,"Number of resolution levels");
  int offEdge  = cimg_option("-offEdeges",1,"Border marge suppression from the data term ");
  bool periodic  = cimg_option("-periodic",false,"periodic boundary conditions (true) else Neumann with zero derivatives (false=default)");
  
  double tol= cimg_option("-tol",1E-6,"precision for CGS trace randomization");
  int nbIt = cimg_option("-nbItTrace",30,"nb iteration trace randomization");
  double precision= cimg_option("-precision",1E-3,"precision for gamma convergence");
  int maxItCount=cimg_option("-maxItConvAlpha",1000,"max number of iteration for alpha's convergence");

  double Umoy  = cimg_option("-Umean",0.,"Average horizontal veocity");
  bool UVinit  = cimg_option("-UVinit",false,"initialisation with a velocity field (foramt uv) ");
  if (UVinit&&nb_resol!=1){nb_resol=1;cout<<"setting a single resolution level!"<<endl;}
  double alphaInit  = cimg_option("-alphaInit",10.,"Horn & Schunck regularization coefficient initialization");
  double betaInit = cimg_option("-betaInit",1.,"Horn & Schunck data term variance inverse initialization");
 
  bool display= cimg_option("-display",false,"display mode");

  int dxOrig=dx;
  int dyOrig=dy;
  int MaxAll;
  if (dxOrig<dyOrig){MaxAll=(int)(dxOrig/2.);}
  else{MaxAll=(int)(dyOrig/2.);}
  double FactorRobParam=1;
  if (norm==2){FactorRobParam=1.;}
  if (norm==1){FactorRobParam=1./(2*sigma_obs);}
  if (norm==0){FactorRobParam=1./(sigma_obs*sigma_obs);}
  double FactorRobParam_reg=1;
  if (normRegRob==2){FactorRobParam_reg=1.;}
  if (normRegRob==1){FactorRobParam_reg=1./(2*sigma_reg);}
  if (normRegRob==0){FactorRobParam_reg=1./(sigma_reg*sigma_reg);}
  double alpha=alphaInit,alpha_tmp=alphaInit;
  double beta=betaInit;
  double sigma_obs_tmp=sigma_obs,sigma_reg_tmp=0,FactorRobParamInit=FactorRobParam; 
  double *MLestimates=new double [4];

    /*Reading images and creation of motion compensate image by mean field*/
    char **fileIm= alloc_seqPath(7);
    char ch[100];
    if (UVinit){read_path(ifn,5,fileIm);}
    else{
       read_path(ifn,4,fileIm);
    } 
     CImg <double> imOrig(dxOrig,dyOrig); 	 
    CImg <double> imdtOrig(dxOrig,dyOrig); 
    CImg <double> imOrigLarge(Dx,2*Dy);
    CImg <double> tmpLarge(Dx,Dy);
    CImg <double> Umean(Dx,Dy); 
    CImg <double> XOrig(dxOrig,dyOrig,2); 	
    CImg <double> mask(dxOrig,dyOrig);
    mask.fill(1.);
    Umean.fill(Umoy);if (UVinit){cout<<fileIm[4]<<endl;}

      
	imOrig=imOrigLarge.get_load_tiff(fileIm[0]).get_slice(0).get_channel(0).get_crop(ofx,ofy,ofx+dxOrig-1,ofy+dyOrig-1);
	if (Umoy!=0){Recal_spline(tmpLarge ,imOrigLarge.get_load_tiff(fileIm[1]).get_slice(0).get_channel(0).get_crop(0,0,Dx-1,Dy-1),fill(Umean,Umean*0.),3);}
	else{
	  if (UVinit){Recal_spline(tmpLarge,imOrigLarge.get_load_tiff(fileIm[1]).get_slice(0).get_channel(0).get_crop(0,0,Dx-1,Dy-1),read_aUVFile(fileIm[4],Dx,Dy,Dx,Dy,ofx,ofy),3);}
	  else{tmpLarge=imOrigLarge.get_load_tiff(fileIm[1]).get_slice(0).get_channel(0).get_crop(0,0,Dx-1,Dy-1);}
	}
	imdtOrig=tmpLarge.get_crop(ofx,ofy,ofx+dxOrig-1,ofy+dyOrig-1);
	

    /*Expanding images for multiresolution*/
    expand_dim(imOrig.width(),imOrig.height(),nb_resol,&dx,&dy);
    cout<<"image size before "<<dxOrig<<" "<<dyOrig;
    cout<<", and after expanding "<<dx<<" "<<dy<<endl;
    CImg <double> im(dx,dy); 
    CImg <double> imdt(dx,dy); 
    
    im=expand(imOrig,nb_resol);
    imdt=expand(imdtOrig,nb_resol); 
    if (display){
      im.display("image 1");
      imdt.display("image 2");
    }
     /*Memory allocation */
    CImgList<double> pyrIm=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrImdt=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrOut=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrMask=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_x=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_y=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_t=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_norm=alloc_pyramid (dx,dy,1,nb_resol);

    CImg <double> X(dx,dy,2); 
    CImg <double> trueUV(dxOrig,dyOrig,2); 
    CImgList<double> pyrXtil=alloc_pyramid (dx,dy,2,nb_resol);
    CImgList<double> pyrXinc=alloc_pyramid (dx,dy,2,nb_resol);
    CImgList<double> pyrXinctmp=alloc_pyramid (dx,dy,2,nb_resol);
    CImgList<double> pyrTmp=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrDiv=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrDiv_x=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrDiv_y=alloc_pyramid (dx,dy,1,nb_resol);
	
 		           	     
    /*-- Construction de la pyramid multiresolution  (image sparse ou non) --*/
    CImg<double> PB_burt(5);PB_burt(0,0)=0.05;PB_burt(1,0)=0.25; PB_burt(2,0)=0.4;PB_burt(3,0)=0.25; PB_burt(4,0)=0.05;
    pyrIm[0]=im;pyrImdt[0]=imdt;pyrMask[0]=expand(mask,nb_resol);
    if (display){pyrIm[0].display();pyrImdt[0].display();}
    for(int resol=0;resol<nb_resol-1;resol++){ 
      pyrMask[resol+1]=Filtre_2D_sparse(Filtre_2D_sparse(pyrMask[resol],PB_burt),PB_burt.get_transpose()).get_resize(pyrIm[resol+1],1,1);
      pyrIm[resol+1]=Filtre_2D_sparse(Filtre_2D_sparse(pyrIm[resol],PB_burt),PB_burt.get_transpose()).get_resize(pyrIm[resol+1],1,1);
      pyrImdt[resol+1]=Filtre_2D_sparse(Filtre_2D_sparse(pyrImdt[resol],PB_burt),PB_burt.get_transpose()).get_resize(pyrImdt[resol+1],1,1);
      if (display){pyrIm[resol+1].display();pyrImdt[resol+1].display();}
    }
    pyrXtil[nb_resol-1].fill(0.);
    if (UVinit){pyrXtil[0]=read_aUVFile(fileIm[4],dx,dy,Dx,Dy,ofx,ofy);}
    /* -- Multiresolution --*/
    for (int resol=nb_resol-1; resol>=0;resol--){
      cout<<"****************************************************************"<<endl;
      cout<<"Resolution: "<<resol<<endl;
      cout<<"****************************************************************"<<endl;
       
	pyrf_x[resol]=gradSpline_X(pyrImdt[resol],sigBlur);
	pyrf_y[resol]=gradSpline_Y(pyrImdt[resol],sigBlur);
	pyrf_t[resol]=grad_t(pyrIm[resol],pyrImdt[resol],sigBlur);
      
    pyrf_norm[resol].fill(1.);
      if (resol==0){
	pyrf_x[resol].div(pyrf_norm[resol]);
	pyrf_y[resol].div(pyrf_norm[resol]);
	pyrf_t[resol].div(pyrf_norm[resol]);
      }
      pyrOut[resol].fill(1.);
      setZeroBorder(pyrOut[resol],offEdge,offEdge);if (resol==0){setZeroBorderOnExpandedRegion(pyrOut[resol],dxOrig,dyOrig);}
      pyrXinc[resol].fill(0.);
       //Horn and Schunck on each resolution level with Bayesian estimation of alpha and beta
      bool conv=false;
      double currentEvidence1,currentEvidenceMin1=1E30; 
      double currentEvidence2,currentEvidenceMin2=1E30;
      double countConv=0;
      int count=0;
      double epsilon;
      alpha=alphaInit;
      beta=betaInit;
      FactorRobParam=FactorRobParamInit;
      CImgList<double> reg_out(2,pyrf_x[resol].width(),pyrf_x[resol].height());
      reg_out[0].fill(1.);reg_out[1].fill(1.);
      if (normRegRob==0){sigma_reg=1./sqrt(FactorRobParam_reg);}
      if (norm==0){sigma_obs=1./sqrt(FactorRobParam);}
      if (norm==1){sigma_obs=1./(2*FactorRobParam);}

      while (!conv && count<Max((int)((double)maxItCount/pow(100.,(double)resol)),1)){
	if (norm==1){pyrOut[resol]=RobustDataCoef_QuasiTV(pyrXinc[resol].get_fill(0.),pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);}
	if (norm==0){pyrOut[resol]=RobustDataCoef(pyrXinc[resol].get_fill(0.),pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);}
	HornAndSchunck(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],alpha,pyrOut[resol],reg_out,periodic);
	if (display){pyrOut[resol].display();(pyrXinc[resol]+pyrXtil[resol]).display();}
	if (norm!=2|| normRegRob!=2){
	   for (int kk=0;kk<5;kk++){ 
	     if (norm==1){pyrOut[resol]=RobustDataCoef_QuasiTV(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);}
	     if (norm==0){pyrOut[resol]=RobustDataCoef(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);}
	     setZeroBorder(pyrOut[resol],offEdge,offEdge); if (resol==0){setZeroBorderOnExpandedRegion(pyrOut[resol],dxOrig,dyOrig);}
	     if(normRegRob==0){UpdateRobustWeight_Leclerc(pyrXinc[resol]+pyrXtil[resol],reg_out,FactorRobParam_reg, periodic) ;}
	     if(normRegRob==1){UpdateRobustWeight_QuasiTV(pyrXinc[resol]+pyrXtil[resol],reg_out,FactorRobParam_reg, periodic) ;}
	     HornAndSchunck(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],alpha,pyrOut[resol],reg_out,periodic);
	   }
	}
	if ( !(sigmaRobEstim || sigmaRegRobEstim)||resol!=0){
	  alpha_tmp=alpha;
	  if (sigmaBetaEstim){
	    beta=evidence_betaHS(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],reg_out,alpha,nbIt,tol,1, periodic);  
	  }
	  if (sigmaAlphaEstim){
	    alpha=evidence_alphaHS(pyrXinc[resol]+pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],reg_out,alpha,beta,nbIt,tol,1, periodic);
	  }
	}
	else{
	  if (sigmaRegRobEstim && normRegRob==0){  
	    MLestimates=evidence_alphaBetaAndTauDataAndRegHS_normalized(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],reg_out,alpha,beta,nbIt,tol,FactorRobParam,FactorRobParam_reg, periodic);
	  }
	  else{
	    if (norm==0){
	      MLestimates=evidence_alphaBetaAndTauHS_normalized(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],reg_out,alpha,beta,nbIt,tol,FactorRobParam, periodic);
	    }
	    if (norm==1){
	      MLestimates=evidence_alphaBetaAndTauHS_normalized_QuasiTV(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],reg_out,alpha,beta,nbIt,tol,FactorRobParam, periodic);
	    }
	  }
	  if (sigmaAlphaEstim){
	   	alpha_tmp=alpha;
		alpha=MLestimates[0];
	  }
	  else { cout<<"(No alpha estimation)"<<endl;alpha_tmp=alpha;}
	  if (sigmaBetaEstim){
	        beta=MLestimates[1];
	  }
	  else { cout<<"(No beta estimation)"<<endl;}
	  if (sigmaRobEstim && norm!=2){
	    if (sqrt((alpha_tmp-alpha)*(alpha_tmp-alpha))< 0.1*alpha){
	    //cout<<"(Sigma Robust estimation)"<<endl;
	      FactorRobParam=MLestimates[2];   
	      sigma_obs_tmp=sigma_obs;
	      if (norm==0){sigma_obs=1./sqrt(FactorRobParam);}
	      if (norm==1){sigma_obs=1./(2*FactorRobParam);}
	    }
	  }
	  if (sigmaRegRobEstim && normRegRob==0){
	    if (sqrt((alpha_tmp-alpha)*(alpha_tmp-alpha))< 0.1*alpha){
	      //if (sqrt((sigma_obs_tmp-sigma_obs)*(sigma_obs_tmp-sigma_obs))< 0.01*sigma_obs){
		FactorRobParam_reg=MLestimates[3];   
		sigma_reg_tmp=sigma_reg;
		sigma_reg=1./sqrt(FactorRobParam_reg);
		//}
	    }
	  }
	  
	}
	epsilon= precision*alpha;
	conv=(sqrt((alpha_tmp-alpha)*(alpha_tmp-alpha))<epsilon);
	if (norm!=2&&sigmaRobEstim&&resol==0){ conv =conv &&(sqrt((sigma_obs_tmp-sigma_obs)*(sigma_obs_tmp-sigma_obs))<precision*sigma_obs) &&  (sqrt((sigma_obs_tmp-sigma_obs)*(sigma_obs_tmp-sigma_obs))!=0);}
	if (sigmaRegRobEstim&& normRegRob==0&&resol==0){ conv =conv &&(sqrt((sigma_reg_tmp-sigma_reg)*(sigma_reg_tmp-sigma_reg))<precision*sigma_reg) &&  (sqrt((sigma_reg_tmp-sigma_reg)*(sigma_reg_tmp-sigma_reg))!=0);}
	cout<<"alpha="<<alpha<<"  (alpha',beta)=("<<alpha*beta<<","<<beta<<"); sigma_d= "<<sigma_obs<<" (factorRobParam_d= "<<FactorRobParam<<")"<<"; sigma_r= "<<sigma_reg<<" (factorRobParam_r= "<<FactorRobParam_reg<<")"<<endl;
  
	count++;
	currentEvidence1=evidence_modelHS_Mackay(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],reg_out,beta,alpha,1,periodic,1);
	currentEvidence2=evidence_modelHS_Mackay(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],reg_out,beta,alpha,1,periodic,2);
	if (currentEvidence1>currentEvidenceMin1 || currentEvidence2>currentEvidenceMin2 && count>2){countConv++;}
	else{
	  countConv=0;
	  if (currentEvidence1<currentEvidenceMin1) {currentEvidenceMin1=currentEvidence1;}
	  if (currentEvidence2<currentEvidenceMin2) {currentEvidenceMin2=currentEvidence2;}
	}
	if (countConv>=3){conv=true;cout << "CONVERGENCE "<<countConv<<" ... "<<endl;}
      } 
      if (display){
	(pyrXinc[resol]+pyrXtil[resol]).get_slice(0).display("U"); (pyrXinc[resol]+pyrXtil[resol]).get_slice(1).display("V");
	(pyrOut[resol]).display("obs");reg_out.display("reg");
      }
      if(resol!=0){  
	Proj_Dup_Spline(pyrTmp[resol-1], (pyrXtil[resol].get_slice(0)+pyrXinc[resol].get_slice(0))*2.,3); 
	FillChannel(pyrXtil[resol-1],pyrTmp[resol-1],0);
	Proj_Dup_Spline(pyrTmp[resol-1], (pyrXtil[resol].get_slice(1)+pyrXinc[resol].get_slice(1))*2.,3); 
	FillChannel(pyrXtil[resol-1],pyrTmp[resol-1],1);
	X=pyrXtil[resol-1].get_resize(X.width(),X.height(),2)*pow(2.,(double)(resol-1.)); 
	//X.get_crop(0,0,dxOrig-1,dyOrig-1).display("UV");
	pyrTmp[resol-1]=pyrImdt[resol-1];
	Recal_spline(pyrImdt[resol-1],pyrTmp[resol-1],pyrXtil[resol-1],3); 
      }
      else{
          
	  X=pyrXtil[0]+pyrXinc[0];
	  sprintf(ch,"%sUV_HS.motion",fileIm[2]);
	  write_aMotionFile(ch,X.get_crop(0,0,dxOrig-1,dyOrig-1));
	  sprintf(ch,"%sUV_HSDataOut.pgm",fileIm[2]);
	  write_pgm255(pyrOut[0].get_crop(0,0,dxOrig-1,dyOrig-1),ch,0,1);	
	  sprintf(ch,"%sUV_HSRegOutX.pgm",fileIm[2]);
	  write_pgm255(reg_out[0].get_crop(0,0,dxOrig-1,dyOrig-1),ch,0,1);	
	  sprintf(ch,"%sUV_HSRegOutY.pgm",fileIm[2]);
	  write_pgm255(reg_out[1].get_crop(0,0,dxOrig-1,dyOrig-1),ch,0,1);	
	  if (display){
	    X.get_crop(0,0,dxOrig-1,dyOrig-1).display("UV");
	    trueUV.display("truth");
	  }
	
      }
    }
    delete [] MLestimates;

    cout<<"\n ** End algorithm **"<<endl;	
    return 0;
  }

















